/* ESA (C) 2000-2018 
*
* This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
*
*    SAS is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    SAS is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with SAS.  If not, see <http:*www.gnu.org/licenses/>.
*/
/*=========================================================================*/
/* ------------------------------------------------------------------------
   edu_simu.c

   Version|By  Date    | Comments
   -------------------------------------------------------------------
   0.0    |ml  14/03/96| Creation 
   0.1    |    19/03/96| (Demain c'est le printemps)
          |            | Ajout des verif. ligne>1024,trop d'evts.. 
   0.2    |    25/03/96| Emission d'un faux evenement pour signaler TIME BLOC
          |            | (mode timing only)     
   0.3    |    25/06/96| Comparaison seuil, mise de :  if( x >= seuil )
   0.4    |    26/06/96| EOL mode TR au bit 12 (et non 13 comme SDD)     
   0.5    |    08/08/96| Mode TIM Mem[] >= th au lieu >
   1.0    |    03/09/97| Nouvelle regle en cas de pixel egaux  pour
          |            | determination du pixel central.
   2.0    |    12/02/98| Introduction position fenetre x0, y0 dans
          |            | coordonnees des evenements EDU.   
   2.1    |    19/02/98| Test sinter >= threshold avec conversion type int
   3.0    |    27/07/98| Modif pour passer sans warning avec :
          |            | gcc  -c -g -Wall -W -Werror edu_simu.c  et lint
   3.1    |    11/12/98| Correction bug: decalage offsetY de 2 dans section
          |            | 'Start pattern analysis'   
          |            | et note sur remplissage de Mem[][]         

   Aim  :  Simulation of EDU in mode: IMA, TIM or THR 
 

   Arg. :  data    pointer on the u_short EDU input 
           threshold threshold
           offsetX offset array in x                   
           offsetY offset array in y                                   
           EDU_Pat Pattern-Mask definition
           x0, y0  position fenetre en Node coordinate.
           gatti   the terrible gatti
           peduout Pointer on structure receiving EDU output 
           nmaxevt Max number of events that the 
                   structure can stored         
           mode    EDU_IMA, EDU_TIM, EDU_TRA, EDU_THR
            
   Note :  The Output is the structure EDU_OUT.

           The identifier OS9 is used to the OS9 version, if not set
           the version is an Unix one.    


C THIS ROUTINE MAKES ON AN IMAGE IM(IDX,IDY) AN ANALYSIS EQUIVALENT
C TO THE ONE PERFORMED IN THE EDU:
C A) TEST IF THE CENTRAL PIXEL (CURRENT PIXEL) IS ABOVE THRESHOLD
C ISEUIL AND THE BRIGHTEST ONE OF THE CENTRAL 3X3 MATRIX
C B) TEST OF THE ISOLATION OF THE EVENT
C C) DETERMINATION OF THE PATTERN NUMBER IN IPATT
C OF THE EDU PATTERN NUMBER IN INEDU
C AND OF THE SUBSET PATTERN NUMBER ISPAT
C D) CALCULATION OF THE DIFFERENT ASSOCIATED ENERGIES

           This simulation follows the true EDU algorithm, except
           that the computation performed in parallel in true EDU
           are simulated in an optimized 
           way, with loop on patterns and use of 24 bits identifiers 
           (Cf. do_edu_lib0.c) which leads to a strictly equivalent
           result (TO BE VERIFIED -> true up to now).

           Modes IMA, TIM, THR : sortie d'une liste d'evts dans EDU_OUT
           Mode  TRA : l'entree data[] est utilisee comme sortie
           il y a donc ECRASEMENT (le traitement est simplement de 
           supprimer les data avec INV jusqu'a EOF) en gardant les bits EOL.
           On utilise toutefois EDU_OUT.npix pour le nbre de pixels > th
           et EDU_OUT.nevt pour nbre de pixel sortis.                           
           
    

  Rq: Way to fill Mem[][]
  -----------------------  
  The CCD image being  x0, y0, dx=10, dy=8,
  pixel numbered in the order they are read.

            y

            ^
            |80 81 82 83 84 85 86 87 88 89
            |70 71 72 73 74 75 76 77 78 79
            |60 61 62 63 64 65 66 67 68 69
            |50 51 52 53 54 55 56 57 58 59
 |    |     |40 41 42 43 44 45 46 47 48 49
 |y0+3|     |30 31 32 33 34 35 36 37 38 39
 |y0+2|     |20 21 22 23 24 25 26 27 28 29
 |y0+1|     |10 11 12 13 14 15 16 17 18 19 
 |y0  |     |00 01 02 03 04 05 06 07 08 09
 |    |      ----------------> x
 OffsetY[]

  ---------------------------------------------------
 |  |  |  |  |x0|x0|x0|x0|x0| ... 
 |  |  |  |  |  |+1|+2|+3|+4|
  ---------------------------------------------------
  OffseX[]


  Mem[][], after being zeroed, is filled as follows, after
  substraction of the relevant offsets:               
   ---------------------------------------------------
  | 0| 0|20|21|22|23|24|25|26|27|28|29| 0| 0| 0| 0| 
   ---------------------------------------------------
  | 0| 0|10|11|12|13|14|15|16|17|18|19| 0| 0| 0| 0|
   ---------------------------------------------------
  | 0| 0|00|01|02|03|04|05|06|07|08|09| 0| 0| 0| 0|
   ---------------------------------------------------
  | 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 
   ---------------------------------------------------
  | 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0|
   ---------------------------------------------------

  As soon as the pixel 22 is filled, pattern analysis starts.
  The first 5*5 matrix analysed is
     ------------
  | 0| 0|20|21|22|
   --------------
  | 0| 0|10|11|12|
   --------------
  | 0| 0|00|01|02|
   --------------
  | 0| 0| 0| 0| 0| 
   --------------
  | 0| 0| 0| 0| 0|
   --------------

  After the analysis of the 5*5 matrix:
   --------------
  |25|26|27|28|29|
   --------------
  |15|16|17|18|19|
   --------------
  |05|06|07|08|09|
   --------------
  | 0| 0| 0| 0| 0|
   --------------
  | 0| 0| 0| 0| 0|
   --------------
  a vertical shift is performed which leads to
   ---------------------------------------------------
  | 0| 0|30|31|32|33|34|35|36|37|38|39| 0| 0| 0| 0|
   ---------------------------------------------------
  | 0| 0|20|21|22|23|24|25|26|27|28|29| 0| 0| 0| 0| 
   ---------------------------------------------------
  | 0| 0|10|11|12|13|14|15|16|17|18|19| 0| 0| 0| 0|
   ---------------------------------------------------
  | 0| 0|00|01|02|03|04|05|06|07|08|09| 0| 0| 0| 0|
   ---------------------------------------------------
  | 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 0| 
   --------------------------------------------------- 
  
  and so on.



  ------------------------------------------------------------------------ */
/*=========================================================================*/

#include <stdio.h>
#include <strings.h>
#include <string.h>
#ifdef OS9
#include <types.h>
#else
#include <sys/types.h>
#endif
#include "epic_defs.h"
#include "edu_simu.h"
#include "edu_lib.h"
#include "tools.h"



/* Memory area (signed!!) */
static short Mem[EDU_NSID][OFFSIZE] ;

/* Array for id24 computation: the array Valid24[] is relevant to
   the following power of 2 values :
** Don't forget the y inversion in the global definition.
   Cf. do_edu_libx.c (the central pixel, set by definition is not included) 
   
   23 08 09 10 11
   22 07 00 01 12
   21 06 xx 02 13
   20 05 04 03 14
   19 18 17 16 15    
*/
static int Valid24[EDU_NSID][EDU_NSID] =
{
 {  524288, 262144, 131072,  65536, 32768 },
 { 1048576,     32,     16,      8, 16384 },
 { 2097152,     64,      0,      4,  8192 },
 { 4194304,    128,      1,      2,  4096 },	
 { 8388608,    256,    512,   1024,  2048 }
} ;


#define SOURCE "edu_simu"

/*
** Main function calling the subfunction relevant to the mode.
*/

void edu_simu( data, threshold, offsetX, offsetY,
               EDU_Pat, x0, y0, gatti, pedu_out, 
               nmaxevt, mode )
u_short     data[] ;   
u_short     threshold ;
u_short     offsetX[] ;
u_short     offsetY[] ;
EDU_PATTERN EDU_Pat[EDU_NPAT] ;
u_short     x0 ;
u_short     y0 ;
u_short     gatti ;
EDU_OUT     *pedu_out ;
int         nmaxevt ;
EDU_MODE    mode ;
{
int     i,j,k ;
int     ii, jj ;  /* subscript for Mem[] */  
void    matrix_analysis_ima();
void    matrix_analysis_tim();
void    matrix_analysis_thr();
u_short d ;
short   sinter ; /* !! signed */
int     y ;      /*  y pixel coordinates */
void    time_bloc();

pedu_out->npix = 0 ;
pedu_out->nevt = 0 ;

/* Reinit. de Mem[][] so that the 1st lines are zeroed at each call */
for(j=0;j<EDU_NSID;j++)for(i=0;i<OFFSIZE;i++) Mem[j][i] = 0 ;

if( mode == EDU_TRA )
{
int kout = 0 ; 
int d1 ;
	
 k = 0 ; 
 y = 2 ; 
 do
 {
 ii = 2 ; 	
   do
  {
   d = data[ k++ ];  /* new pixel */
   if( !(d & SE_INV) )
   {
    d1 = (d&0xfff)-gatti-offsetX[x0+ii-2]-offsetY[y0+y-2] ;
    if( d1 >= (int)threshold ) pedu_out->npix++ ;
    d1 &= 0xfff ;  /* Mise sur 12 bits */
    if( d & SE_EOL ) d1 |= HBR_EOL ;  /* Repport du EOL */
    ii++ ;
    data[ kout++ ] = d1 ;
   }
  } while( !(d & SE_EOL) ) ;
  y++ ;
 } while( !(d & SE_EOF) ) ;

 pedu_out->nevt = kout ;

	
}
else
{
 /*== Data reception with gatti and offsets substraction ==*/
 /* Memorisation de 2 lignes avant debut analyse pattern */
 ii = jj = 2 ;
 k = 0 ; 
 do
 {
  do
  { 
   d = data[ k++ ];
   if( !(d & SE_INV) ) 
   {
    sinter = (d&0xfff)-gatti-offsetX[x0+ii-2]-offsetY[y0+jj-2] ;
    if( (int)sinter >= (int)threshold ) pedu_out->npix++ ;
    Mem[ jj ][ ii++ ] = sinter ;
   }
   if(ii>=OFFSIZE) {printf("NO EOL AFTER %d DATA\n",OFFSIZE);  return; }
  } while( !(d & SE_EOL) );
  jj++; ii=2;
 } while( jj < 4 );


 /*= Start pattern analysis =*/
 /* in what follows jj stays at 4, since we only fill the last line */
 y = 2 ; 
 do
 {
  ii = 2 ;	
  do /* 2 nouveaux pixels */
  { 
   d = data[ k++ ];
   if( !(d & SE_INV) )
   { 
    sinter = (d&0xfff)-gatti-offsetX[x0+ii-2]-offsetY[y0+y] ;
    if( (int)sinter >= (int)threshold ) pedu_out->npix++ ;
    Mem[ 4 ][ ii++ ] = sinter ;
   }
   if(ii>=OFFSIZE) {printf("NO EOL AFTER %d DATA\n",OFFSIZE);  return; }
  } while( ii < 4 ); 
  
  do
  {
   d = data[ k++ ];  /* new pixel */
   if( !(d & SE_INV) )
   {
    sinter = (d&0xfff)-gatti-offsetX[x0+ii-2]-offsetY[y0+y] ;
    if( (int)sinter >= (int)threshold ) pedu_out->npix++ ;
   	Mem[ 4 ][ ii++ ] = sinter ;
    /* Matrix analysis */
    switch( mode )
    {
    case EDU_IMA : 	
      matrix_analysis_ima(ii-3, y, threshold, EDU_Pat, pedu_out, x0, y0);
      break ;
    case EDU_TIM : 
      matrix_analysis_tim(ii-3, y, threshold, EDU_Pat, pedu_out, x0, y0);
      break ;
    case EDU_THR :   
      matrix_analysis_thr(ii-3, y, threshold, EDU_Pat, pedu_out, x0, y0);
      break ;
    case EDU_TRA : /* irrelevent */
      break ; 
    }
    if(ii>=OFFSIZE) {printf("NO EOL AFTER %d DATA\n",OFFSIZE);  return; }
    if( pedu_out->nevt >= (u_int)nmaxevt ) return ; /* stop if too much evts */ 
   }
  } while( !(d & SE_EOL) ) ;
  
  /*AI: Centos 7 does not process this memcpy sentence properly.
   *The documentation of memcpy warns about the usage of this 
   * directive in overlapping memory blocks. Therefore, we change it to 
   * memmove
   */
  /* decalage lignes [1,4] -> [0,3] */
  memmove( (char *)&Mem[0][0], (char *)&Mem[1][0], 
          sizeof(u_short)*(EDU_NSID-1)*OFFSIZE   ); 
  y++ ;
  
  /* En mode timing, tous les 1024 lignes, on doit
  ** generer un bloc TIME (Cf. edu_format.c), pour
  ** memoriser l'endroit ou mettre ce bloc dans la
  ** liste d'evts, on genere ici un evt ou x=y=-1
  */
  if( mode == EDU_TIM )
  if( (0<y) && !(y % 1024) ) time_bloc(pedu_out,nmaxevt); 
 
 } while( !(d & SE_EOF) ) ;
  
} 

} /*-- end of edu_simu_ima() -----------------------*/


/* Ajout FAlSE event to indique the location of a TIME BLOC 
** x=y=0xffff is used to indicate that it is a time bloc.
** The number of pixels above threshold in the pseudo-frame
** is stored in ncp
*/
void  time_bloc(pedu_out, nmaxevt)
EDU_OUT     *pedu_out ;
int         nmaxevt ;
{
  pedu_out->Evt[ pedu_out->nevt ].x  = 0xffff ;
  pedu_out->Evt[ pedu_out->nevt ].y  = 0xffff  ;       
  pedu_out->Evt[ pedu_out->nevt ].n  = 0 ;
  pedu_out->Evt[ pedu_out->nevt ].cp = 0 ;
  pedu_out->Evt[ pedu_out->nevt ].e2 = 0 ;
  pedu_out->Evt[ pedu_out->nevt ].e3 = 0 ;
  pedu_out->Evt[ pedu_out->nevt ].e4 = 0 ;
  pedu_out->Evt[ pedu_out->nevt ].ncp= pedu_out->npix   ;
  /* reset npix */ 
  pedu_out->npix = 0 ; 
 
  pedu_out->nevt++ ;
  if( pedu_out->nevt >= (u_int)nmaxevt ) return ; /* stop if too much evts */ 
}


void matrix_analysis_ima( ic, y, thre, EDU_Pat, pedu_out, x0, y0 ) 
int          ic;  /*  x coordinate of central pixel of the 5*5 matrix 
                      in Mem[][ic] = x coordinate of the pattern      */
int          y ;  /*  y coordinate of the pattern */
int          thre ;
EDU_PATTERN  EDU_Pat[EDU_NPAT] ;
EDU_OUT      *pedu_out ;
u_short      x0 ;
u_short      y0 ;
{
int      cp ;
int      val[8] ;
int      ii, jj ;
int      ener2,ener3,ener4 ;
int      i ;
u_short  ncp ;
int      np ;
BOOLEAN  pattern_found ;
int      id24_data ;


 cp = Mem[2][ic] ;  /* central pixel */	
 if( cp >= thre ) 
 {    
 /*       	
 C TEST IF CENTRAL PIXEL IS THE BRIGHTEST ONE (in 3*3 matrix)
 C WE FOLLOW A WELL DEFINED WAY (TO BE CHECKED WITH THE EDU CHOICE)
 C The central pixel is considered as the brightest is the following
 C conditions are true:
    x  x  x  x  x 
    x <=  <  <  x
    x <=  cp <  x
    x <=  <= <  x
    x  x  x  x  x
 C 
 C Order of testing (array val[8]) :
 C  x x x x x
 C  x 8 1 2 x
 C  x 7 . 3 x
 C  x 6 5 4 x
 C  x x x x x
 C
 */ 
 val[0]=Mem[3][ic] ;
 if(val[0] < cp) 
 {
  val[1]=Mem[3][ic+1];
  if(val[1] < cp) 
  {
   val[2]=Mem[2][ic+1];
   if(val[2] < cp) 
   {
    val[3]=Mem[1][ic+1];
    if(val[3] <  cp) 
    {
     val[4]=Mem[1][ic];
     if(val[4] <= cp) 
     {
      val[5]=Mem[1][ic-1];
      if(val[5] <= cp) 
      {
       val[6]=Mem[2][ic-1];
       if(val[6] <= cp) 
       {
       val[7]=Mem[3][ic-1]; 
       if(val[7] <= cp) 
       {
       /*	
       ** Check if the 5*5 matrix is relevant to a pattern, for pixels
       ** within the mask.
       */ 
       /* calcul id24 */  
       id24_data = 0 ;
       for(jj=0;jj<EDU_NSID;jj++)
       for(ii=0;ii<EDU_NSID;ii++)
          if( Mem[jj][ic-2+ii] >= thre ) id24_data += Valid24[jj][ii] ; 

       pattern_found = FALSE ;
       for( np=0; np<EDU_NPAT; np++ )
       {
       int mask = EDU_Pat[np].id_mask ;

        /* Comparaison id reconstruit masque, avec celui de la bibliotheque. */ 	
       	if( (EDU_Pat[np].id_patt & mask) == (id24_data & mask) )
        { pattern_found = TRUE;  break; }
                 
       }
       if( pattern_found == FALSE ) return ;
    
       /* Energies computation */  
       ener2=ener3=0 ;
       for(i=0;i<8;i++)        
        if(val[i] >= thre) ener2 += val[i] ;
        else               ener3 += val[i] ;
      
       /*  ener4 = SUM OF THE CORONNA PIXELS < THRESHOLD
       **  ncp   = Number of pixels >= threshold in coronna */ 
       ener4= 0 ;  ncp  = 0 ;
       for( i=0; i<5; i++ )
       { int inter = Mem[ 0 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic+2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=0; i<5; i++ )
       { int inter = Mem[ 4 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic-2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       
       /*
       C CENTRAL PIXEL ENERGY = cp
       C INSIDE THE INNER 3X3 MATRIX WITHOUT CENTRAL PIXEL :
       C ENERGY >= thre :  ener2 15 bits non signes [0,32767]
       C ENERGY <  thre :  ener3 14 bits signes [-16384,+16383]
       C INDSIDE THE CORONNA :
       C PIXELS < thre = ener4  14 bits signes [-16384,+16383]
       C ncp = Number of pixels >= thre in coronna ( saturation a 7 (3 bits) ) 
       C EDU PATTERN NUMBER = np
       */ 
       /* Saturations : */ 
       if( (ener3<-16384)||(16383<ener3) ) ener3 = -16384 ;
       if( (ener4<-16384)||(16383<ener4) ) ener4 = -16384 ;
       if( ncp > 7 ) ncp = 7 ;

       pedu_out->Evt[ pedu_out->nevt ].x  = ic + x0 ;
       pedu_out->Evt[ pedu_out->nevt ].y  = y  + y0 ;       
       pedu_out->Evt[ pedu_out->nevt ].n  = np ;
       pedu_out->Evt[ pedu_out->nevt ].cp = cp ;
       pedu_out->Evt[ pedu_out->nevt ].e2 = ener2 ;
       pedu_out->Evt[ pedu_out->nevt ].e3 = ener3 ;
       pedu_out->Evt[ pedu_out->nevt ].e4 = ener4 ;
       pedu_out->Evt[ pedu_out->nevt ].ncp= ncp   ;
   
       pedu_out->nevt++ ;
       
       }
       }
      }
     }
    }
   }
  }
 }
 }

} /*---------------------------------------*/


void matrix_analysis_tim( ic, y, thre, EDU_Pat, pedu_out, x0, y0 ) 
int          ic;  /*  x coordinate of central pixel of the 5*5 matrix 
                      in Mem[][ic] = x coordinate of the pattern      */
int          y ;  /*  y coordinate of the pattern */
int          thre ;
EDU_PATTERN  EDU_Pat[EDU_NPAT] ;
EDU_OUT      *pedu_out ;
u_short      x0 ;
u_short      y0 ;
{
int      cp ;
int      val[8] ;
int      ii, jj ;
int      ener2,ener3,ener4 ;
int      i ;
u_short  ncp ;
int      np ;
BOOLEAN  pattern_found ;
int      id24_data ;


 cp = Mem[2][ic] ;  /* central pixel */	
 if( cp >= thre ) 
 {    
 

 /* In timing no TEST IF THE CENTRAL PIXEL IS THE BRIGHTEST ONE */
 /* We keep val[] init. for e3 computation */
 val[0]=Mem[3][ic] ;
 val[1]=Mem[3][ic+1];
 val[2]=Mem[2][ic+1];
 val[3]=Mem[1][ic+1];
 val[4]=Mem[1][ic];
 val[5]=Mem[1][ic-1];
 val[6]=Mem[2][ic-1];
 val[7]=Mem[3][ic-1]; 

 /************************ 
 val[0]=Mem[3][ic] ;
 if(val[0] < cp) 
 {
  val[1]=Mem[3][ic+1];
  if(val[1] < cp) 
  {
   val[2]=Mem[2][ic+1];
   if(val[2] < cp) 
   {
    val[3]=Mem[1][ic+1];
    if(val[3] <  cp) 
    {
     val[4]=Mem[1][ic];
     if(val[4] <= cp) 
     {
      val[5]=Mem[1][ic-1];
      if(val[5] <= cp) 
      {
       val[6]=Mem[2][ic-1];
       if(val[6] <= cp) 
       {
       val[7]=Mem[3][ic-1]; 
       if(val[7] <= cp) 
       {
       	
********************/       
       /*	
       ** Check if the 5*5 matrix is relevant to a pattern, for pixels
       ** within the mask.
       */ 
       /* calcul id24 */  
       id24_data = 0 ;
       for(jj=0;jj<EDU_NSID;jj++)
       for(ii=0;ii<EDU_NSID;ii++)
          if( Mem[jj][ic-2+ii] >= thre ) id24_data += Valid24[jj][ii] ; 

       pattern_found = FALSE ;
       for( np=0; np<EDU_NPAT; np++ )
       {
       int mask = EDU_Pat[np].id_mask ;

        /* Comparaison id reconstruit masque, avec celui de la bibliotheque. */ 	
       	if( (EDU_Pat[np].id_patt & mask) == (id24_data & mask) )
        { pattern_found = TRUE;  break; }
                 
       }
       if( pattern_found == FALSE ) return ;
    
       /* Timing energies computation */
       ener2 = Mem[2][ic+1] ;
       if( ener2 < thre ) ener2 = 0 ;  

       /* Energies computation !!!! TBC !!!!! : ONLY ener1 and ener2 is relevant anyway */  
       ener3=0 ;
       for(i=0;i<8;i++)        
        if(val[i] < thre) ener3 += val[i] ;
      
       /*  ener4 = SUM OF THE CORONNA PIXELS < THRESHOLD
       **  ncp   = Number of pixels >= threshold in coronna */ 
       ener4= 0 ;  ncp  = 0 ;
       for( i=0; i<5; i++ )
       { int inter = Mem[ 0 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic+2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=0; i<5; i++ )
       { int inter = Mem[ 4 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic-2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       
       /*
       C CENTRAL PIXEL ENERGY = cp
       C INSIDE THE INNER 3X3 MATRIX WITHOUT CENTRAL PIXEL :
       C ENERGY >= thre :  ener2 15 bits non signes [0,32767]
       C ENERGY <  thre :  ener3 14 bits signes [-16384,+16383]
       C INDSIDE THE CORONNA :
       C PIXELS < thre = ener4  14 bits signes [-16384,+16383]
       C ncp = Number of pixels >= thre in coronna ( saturation a 7 (3 bits) ) 
       C EDU PATTERN NUMBER = np
       */ 
       /* Saturations : */ 
       if( (ener3<-16384)||(16383<ener3) ) ener3 = -16384 ;
       if( (ener4<-16384)||(16383<ener4) ) ener4 = -16384 ;
       if( ncp > 7 ) ncp = 7 ;

       /*
       C CENTRAL PIXEL ENERGY = cp
       C INSIDE THE INNER 3X3 MATRIX WITHOUT CENTRAL PIXEL :
       C ENERGY >= thre :  ener2 15 bits non signes [0,32767]
       C ENERGY <  thre :  ener3 14 bits signes [-16384,+16383]
       C INDSIDE THE CORONNA :
       C PIXELS < thre = ener4  14 bits signes [-16384,+16383]
       C ncp = Number of pixels >= thre in coronna ( saturation a 7 (3 bits) ) 
       C EDU PATTERN NUMBER = np
       */ 
       /* Pas de calcul de saturation  */

       pedu_out->Evt[ pedu_out->nevt ].x  = ic + x0 ;
       pedu_out->Evt[ pedu_out->nevt ].y  = y  + y0 ;       
       pedu_out->Evt[ pedu_out->nevt ].n  = np ;
       pedu_out->Evt[ pedu_out->nevt ].cp = cp ;
       pedu_out->Evt[ pedu_out->nevt ].e2 = ener2 ;
       pedu_out->Evt[ pedu_out->nevt ].e3 = ener3 ;
       pedu_out->Evt[ pedu_out->nevt ].e4 = ener4 ;
       pedu_out->Evt[ pedu_out->nevt ].ncp= ncp   ;
   
       pedu_out->nevt++ ;
/*****************       
       }
       }
      }
     }
    }
   }
  }
 }  **********/
 }

} /*---------------------------------------*/


void matrix_analysis_thr( ic, y, thre, EDU_Pat, pedu_out, x0, y0 ) 
int          ic;  /*  x coordinate of central pixel of the 5*5 matrix 
                      in Mem[][ic] = x coordinate of the pattern      */
int          y ;  /*  y coordinate of the pattern */
int          thre ;
EDU_PATTERN  EDU_Pat[EDU_NPAT] ;
EDU_OUT      *pedu_out ;
u_short      x0 ;
u_short      y0 ;
{
int      cp ;
int      val[8] ;
int      ii, jj ;
int      ener2,ener3,ener4 ;
int      i ;
int      ncp ;
int      np ;
BOOLEAN  pattern_found ;
int      id24_data ;


 cp = Mem[2][ic] ;  /* central pixel */	
 if( cp >= thre ) 
 {    


 /* In THRE no TEST IF THE CENTRAL PIXEL IS THE BRIGHTEST ONE */
 /* We keep val[] init. for e3 computation */
 val[0]=Mem[3][ic] ;
 val[1]=Mem[3][ic+1];
 val[2]=Mem[2][ic+1];
 val[3]=Mem[1][ic+1];
 val[4]=Mem[1][ic];
 val[5]=Mem[1][ic-1];
 val[6]=Mem[2][ic-1];
 val[7]=Mem[3][ic-1]; 

 /************************ 
 val[0]=Mem[3][ic] ;
 if(val[0] < cp) 
 {
  val[1]=Mem[3][ic+1];
  if(val[1] < cp) 
  {
   val[2]=Mem[2][ic+1];
   if(val[2] < cp) 
   {
    val[3]=Mem[1][ic+1];
    if(val[3] <  cp) 
    {
     val[4]=Mem[1][ic];
     if(val[4] <= cp) 
     {
      val[5]=Mem[1][ic-1];
      if(val[5] <= cp) 
      {
       val[6]=Mem[2][ic-1];
       if(val[6] <= cp) 
       {
       val[7]=Mem[3][ic-1]; 
       if(val[7] <= cp) 
       {
       	
********************/       
       /*	
       ** Check if the 5*5 matrix is relevant to a pattern, for pixels
       ** within the mask.
       */ 
       /* calcul id24 */  
       id24_data = 0 ;
       for(jj=0;jj<EDU_NSID;jj++)
       for(ii=0;ii<EDU_NSID;ii++)
          if( Mem[jj][ic-2+ii] >= thre ) id24_data += Valid24[jj][ii] ; 

       pattern_found = FALSE ;
       for( np=0; np<EDU_NPAT; np++ )
       {
       int mask = EDU_Pat[np].id_mask ;

        /* Comparaison id reconstruit masque, avec celui de la bibliotheque. */ 	
       	if( (EDU_Pat[np].id_patt & mask) == (id24_data & mask) )
        { pattern_found = TRUE;  break; }
                 
       }
       if( pattern_found == FALSE ) return ;
    
   
       /* Energies computation !!!! TBC !!!!! : ONLY ener1 is relevant anyway */  
       /* Timing energies computation */
       ener2 = Mem[2][ic+1] ;
       if( ener2 < thre ) ener2 = 0 ;  

       ener3=0 ;
       for(i=0;i<8;i++)        
        if(val[i] < thre) ener3 += val[i] ;


       /*  ener4 = SUM OF THE CORONNA PIXELS < THRESHOLD
       **  ncp   = Number of pixels >= threshold in coronna */ 
       ener4= 0 ;  ncp  = 0 ;
       for( i=0; i<5; i++ )
       { int inter = Mem[ 0 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic+2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=0; i<5; i++ )
       { int inter = Mem[ 4 ][ ic-2+i ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       for( i=1; i<4; i++ )
       { int inter = Mem[ i ][ ic-2   ] ;	
         if(inter < thre) ener4 += inter; else  ncp++ ;
       }
       
       /*
       C CENTRAL PIXEL ENERGY = cp
       C INSIDE THE INNER 3X3 MATRIX WITHOUT CENTRAL PIXEL :
       C ENERGY >= thre :  ener2 15 bits non signes [0,32767]
       C ENERGY <  thre :  ener3 14 bits signes [-16384,+16383]
       C INDSIDE THE CORONNA :
       C PIXELS < thre = ener4  14 bits signes [-16384,+16383]
       C ncp = Number of pixels >= thre in coronna ( saturation a 7 (3 bits) ) 
       C EDU PATTERN NUMBER = np
       */ 
       /* Saturations : */ 
       if( (ener3<-16384)||(16383<ener3) ) ener3 = -16384 ;
       if( (ener4<-16384)||(16383<ener4) ) ener4 = -16384 ;
       if( ncp > 7 ) ncp = 7 ;



       /*
       C CENTRAL PIXEL ENERGY = cp
       C INSIDE THE INNER 3X3 MATRIX WITHOUT CENTRAL PIXEL :
       C ENERGY >= thre :  ener2 15 bits non signes [0,32767]
       C ENERGY <  thre :  ener3 14 bits signes [-16384,+16383]
       C INDSIDE THE CORONNA :
       C PIXELS < thre = ener4  14 bits signes [-16384,+16383]
       C ncp = Number of pixels >= thre in coronna ( saturation a 7 (3 bits) ) 
       C EDU PATTERN NUMBER = np
       */ 
       /* Pas de calcul de saturation */ 

       pedu_out->Evt[ pedu_out->nevt ].x  = ic + x0 ;
       pedu_out->Evt[ pedu_out->nevt ].y  = y  + y0 ;       
       pedu_out->Evt[ pedu_out->nevt ].n  = np ;
       pedu_out->Evt[ pedu_out->nevt ].cp = cp ;
       pedu_out->Evt[ pedu_out->nevt ].e2 = ener2 ;
       pedu_out->Evt[ pedu_out->nevt ].e3 = ener3 ;
       pedu_out->Evt[ pedu_out->nevt ].e4 = ener4 ;
       pedu_out->Evt[ pedu_out->nevt ].ncp= ncp   ;
   
       pedu_out->nevt++ ;
/*****************       
       }
       }
      }
     }
    }
   }
  }
 }  **********/
 }

} /*---------------------------------------*/

