// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// edusoft.nag.cc
// Interface NAG F90 - EDUSOFT
// ml 14/04/99  C++ version obtained from C previous version edusoft.nag.c
//              Enable to use the SSC standard Error and Message display.
//    20/04/99  Mise des messages en niveau AppLib  


/* ------------------------------------------------------------------------
   Attention ! X0 et Y0 ne servent que pour les offsets, pas pour l'image
  ------------------------------------------------------------------------ */

#include <sys/types.h>

// C++ declarations 
#include <errstr.h>

#include "tools.h"
#include "epic_defs.h"
#include "edu_lib.h"
#include "edu_simu.h"

#define SOURCE "edusoft"

//-- I/F F90-C++ -------
#include "F90cxxGlue.h"
//----------------------


// edu_simu() function declared as a C function:
extern "C" void edu_simu( 
u_short *,   
u_short, 
u_short *,
u_short *,
EDU_PATTERN *,
u_short      ,
u_short      ,
u_short      ,
EDU_OUT * ,
int       ,
EDU_MODE      ) ;


// C++ function calling C edu_simu()
// with use of SSC error messages.
void edusoft_ml( 
int   *mode ,
EDU_PATTERN (*EDU_Pat)[EDU_NPAT] ,
int   *DX, 
int   *DY ,
short *im ,
short *seuileds ,
short *offx, 
short *offy ,
int   *X0, 
int   *Y0 ,
int   *es_nmax ,
EDU_OUT    *EDU_Out )
{
int   i,j,k;


 Sas::message(Msg::AppLib,Msg::Noisy,"edu_simu() call");

 /*===== Call simulation =========*/
 /* Build EMAE-like image with EOL and EOF */
 k = 0 ;
 for( j=0; j<*DY; j++ )
 {
  for(i=0; i<*DX; i++)  *(im + k++) &= 0xfff ;  /* only 12 bits */
  *(im + k - 1 ) |= SE_EOL ;
 }
 *(im + k - 1) |= SE_EOF ;  

 errstr << k << " words in input image" << Sas::message(Msg::AppLib,Msg::Noisy);
 errstr << "Mode number = " << *mode   << Sas::message(Msg::AppLib,Msg::Noisy);

 edu_simu( (u_short *)im, (u_short)*seuileds,(u_short *) offx, (u_short *) offy,
            *EDU_Pat, (u_short)*X0,(u_short)*Y0,(u_short)0, EDU_Out, 
            *es_nmax, (EDU_MODE) *mode );
   
 errstr << (*EDU_Out).nevt << " events found and " << (*EDU_Out).npix << 
   " pixels above threshold." << Sas::message(Msg::AppLib,Msg::Noisy);

 if( (int) (*EDU_Out).nevt == *es_nmax)
 {
 errstr << "Maximum number of events for the used buffer is reached : " << *es_nmax <<  
 Sas::warning("edu_simu 10");
 }

 return ;
}


// === The interface with F90 call ==========================
void CXXFUNC( edusoft, edusoft )
  ( 
   int   *mode ,
   EDU_PATTERN (*EDU_Pat)[EDU_NPAT] ,
   int   *DX, 
   int   *DY ,
   short *im ,
   short *seuileds ,
   short *offx, 
   short *offy ,
   int   *X0, 
   int   *Y0 ,
   int   *es_nmax ,
   EDU_OUT    *EDU_Out
  )
{

edusoft_ml( mode, EDU_Pat, DX, DY, im, seuileds, offx, offy, X0, Y0,
            es_nmax, EDU_Out );

} // edusoft()

// ==========================================================




/* ------------------------------------------------------------------------
   Interface entre F90 et C edu_simu.c with pattern library from the CAL.
  ------------------------------------------------------------------------ */

void pat_init_ml(
char         TABF90[EDU_NSID][EDU_NSID][EDU_NPAT],
EDU_PATTERN  EDU_Pat[EDU_NPAT] 
)
{
int     k, ii, jj;
char    TAB[EDU_NPAT][EDU_NSID][EDU_NSID];
void    fill_edu_pattern( char TAB[EDU_NPAT][EDU_NSID][EDU_NSID], EDU_PATTERN EDU_Pat[] );


  // Pattern number is most quickly varying index in CAL,
  // must be changed to least quickly varying index
  for(k=0;k<EDU_NPAT;k++) 
  for(jj=0;jj<5;jj++) 
  for(ii=0;ii<5;ii++)
     TAB[k][jj][ii] = TABF90[ii][jj][k] ;

  fill_edu_pattern( TAB, EDU_Pat ) ;

  k=0;
  Sas::message(Msg::AppLib,Msg::Noisy,"Pattern  0 : ");
  for(jj=0;jj<5;jj++) 
  errstr <<(0xff & EDU_Pat[k].mat[jj][0])<<" "<<(0xff & EDU_Pat[k].mat[jj][1])<<" "<<
           (0xff & EDU_Pat[k].mat[jj][2])<<" "<<(0xff & EDU_Pat[k].mat[jj][3])<<" "<<
           (0xff & EDU_Pat[k].mat[jj][4])<< Sas::message(Msg::AppLib,Msg::Noisy) ;
  Sas::message(Msg::AppLib,Msg::Noisy,"Pattern 31 : ");
  k=31;
  for(jj=0;jj<5;jj++) 
  errstr <<(0xff & EDU_Pat[k].mat[jj][0])<<" "<<(0xff & EDU_Pat[k].mat[jj][1])<<" "<<
           (0xff & EDU_Pat[k].mat[jj][2])<<" "<<(0xff & EDU_Pat[k].mat[jj][3])<<" "<<
           (0xff & EDU_Pat[k].mat[jj][4])<< Sas::message(Msg::AppLib,Msg::Noisy) ;
 
  return ;
}

void fill_edu_pattern( 
char         TAB[EDU_NPAT][EDU_NSID][EDU_NSID],
EDU_PATTERN  EDU_Pat[]
)
{
int i;
int ii,jj;

 for( i=0; i<EDU_NPAT; i++ )
 {
 register int bi, id ;
 char *pr_c = &TAB[i][0][0] ;

  id = 0 ;
  bi = 1 ;

  id += bi* ( *( pr_c + 17 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 18 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 13 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  8 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  7 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  6 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 11 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 16 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 21 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 22 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 23 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 24 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 19 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 14 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  9 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  4 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  3 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  2 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  1 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  0 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  5 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 10 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 15 )==1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 20 )==1 ? 1 : 0  ) ;

  EDU_Pat[i].id_patt  = id ;

  id = 0 ;
  bi = 1 ;

  id += bi* ( *( pr_c + 17 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 18 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 13 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  8 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  7 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  6 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 11 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 16 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 21 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 22 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 23 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 24 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 19 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 14 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  9 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  4 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  3 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  2 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  1 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  0 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c +  5 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 10 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 15 )>=1 ? 1 : 0  ) ;
  bi *= 2 ;
  id += bi* ( *( pr_c + 20 )>=1 ? 1 : 0  ) ;

  EDU_Pat[i].id_mask  = id ;
     
  /* matrix 5*5 */
  for(jj=0;jj<EDU_NSID;jj++)
   for(ii=0;ii<EDU_NSID;ii++)
      EDU_Pat[i].mat[jj][ii] = TAB[i][jj][ii] ; 


  /*= Number of pixels  =*/
  {
  int ii, jj, nbr=1 ;  /* 1 for the central pixel */
  
    for(ii=0;ii<EDU_NSID;ii++)
    for(jj=0;jj<EDU_NSID;jj++)
    {
      if( (ii==2)&&(jj==2) ) continue ;   /* we skip the central pixel */ 
      if( TAB[i][ii][jj] == 1 ) nbr++ ;
    }
    EDU_Pat[i].number = nbr;   
  } 

  /* le type du pattern  */
  EDU_Pat[i].kind = 'X';

 } /* end of for i */

} /*--- end of fill_edu_pattern() ---------------------*/



//=== The interface with F90 ==================================

void CXXFUNC( pat_init, pat_init )
  (
   char        TABF90[EDU_NSID][EDU_NSID][EDU_NPAT],
   EDU_PATTERN EDU_Pat[EDU_NPAT] 
  )
{

 pat_init_ml( TABF90, EDU_Pat );

} // pat_init()

//=============================================================
